/*
 MacCaml Graphic Library: Mac OS Graphic Library for MacCaml running Caml Light.
 Copyright (C) 2004-2016 Jean Mouric 35700 Rennes France
 email: jean.mouric@orange.fr
 
 This program is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

// File libgraphics.m

#import <Cocoa/Cocoa.h>
#import "MCEvent.h"
#import "MCProtocol.h"
#include "../runtime/fail.h"
#include "../runtime/memory.h"
#include "../runtime/alloc.h" // pour architecture x86_64 (prototype de alloc_tuple())
#include "../runtime/signals.h"
#include "MCGraphicView.h"

#define pi 3.14159265358979323846

/*----------------------------------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------------------------------*/

//Variables globales

#ifdef MACOS
extern NSString* serverName;
#else
NSString* serverName;
#endif

int tailleImage;
int width;
int height;
NSBitmapImageRep* bitmap = nil;
unsigned char* data = nil;
float red, green, blue;
NSRect frame;
NSString* title = nil;
NSTimer* timer = nil;
char fontName[255];
float x;
float y;
float fontSize;
int lineWidth;
BOOL needsRefresh = NO;
BOOL newTitle = NO;
BOOL graphIsOpen = NO;
NSLock* lock = nil;		//synchronisation du thread principal (commandes ralisant le dessin) et du thread
//secondaire, ce dernier envoyant priodiquement ce qui a t dessin au serveur
// qui gre la fentre graphique

NSDistantObject <ServerProtocol>* mcProxy = nil;	// proxy pour la vue vendue par le serveur MacCaml
NSMutableArray* queue = nil; // reproduit la file des vnements en provenance du serveur MacCaml

/*----------------------------------------------------------------------------------------------------*/

//Classe pour un objet grant dans un thread secondaire la relation avec la vue vendue par MacCaml,
//La boucle interactive de Caml fonctionnant dant le thread principal.

@interface MCClient:NSObject
{
    int b;
}
- (void) mcThread;
- (void) refresh:(NSTimer*) timer;
@end

@implementation MCClient
- (id) init
{
    if (self = [super init])
    {
        b=0;
    }
    return self;
}

-(void) refresh:(NSTimer*) tmr
{
    if (b == 1)
    {
        usleep(100000);
        b = 0;
    }
    [lock lock];
    if (graphIsOpen == NO)
    {
        if (data) {free(data); data = nil;}
        bitmap = nil;
        [timer invalidate];
        timer = nil;
        [lock unlock];
        queue = nil;
        return;
    }
    if (newTitle == YES)
    {
        [mcProxy setServerFrame:frame title:title];
        newTitle = NO;
        needsRefresh = YES;
    }
    if (needsRefresh == YES)
    {
        [mcProxy getImage: [NSData dataWithBytes: data length: tailleImage]];
        needsRefresh = NO;
    }
    
    MCEvent* e = nil;
l:	e = [mcProxy getNextEvent];
    if (e == nil)
    {
        [lock unlock];
        return;
    }
    if ([e mask] == WINDOWCLOSED)
    {
        graphIsOpen = NO;
        needsRefresh = NO;
        [mcProxy getImage: nil];
        [lock unlock];
        return;
    }
    if (queue != nil)
    {
        MCEvent* ee = [[MCEvent alloc]init];
        [ee setMask:[e mask]];
        [ee setMouse_x:[e mouse_x]];
        [ee setMouse_y:[e mouse_y]];
        [ee setKey:[e key]];
        [queue addObject:ee];
        [lock unlock];
        usleep(20000);
        [lock lock];
        goto l;
    }
    
    [lock unlock];
}

- (void) mcThread
{
    [lock lock];
    mcProxy = nil;
    mcProxy = (NSDistantObject<ServerProtocol>*) [NSConnection
                                                  rootProxyForConnectionWithRegisteredName:serverName
                                                  host:nil];
    if (mcProxy == nil)
    {
        NSLog(@"Connexion Failure to main thread");
        graphIsOpen = NO;
        [lock unlock];
        return;
    }
    [[mcProxy connectionForProxy] enableMultipleThreads];
    
    [(NSDistantObject*)mcProxy setProtocolForProxy:@protocol(ServerProtocol)];
    [mcProxy setServerFrame:frame title:title];
    
    b = 1;
    
    timer = [NSTimer scheduledTimerWithTimeInterval:0.1 target:self selector:@selector(refresh:)
                                           userInfo:nil repeats:YES];
    
    [[NSRunLoop currentRunLoop] addTimer:timer forMode:NSDefaultRunLoopMode];
    [lock unlock];
    [[NSRunLoop currentRunLoop]run];
}
@end

/*----------------------------------------------------------------------------------------------------*/

//Les fonctions graphiques de Caml

MCClient* mcClient = nil; //l'objet charg de la relation avec MacCaml

void grfail(fmt, arg)
char * fmt, * arg;
{
    char buffer[1024];
    sprintf(buffer, fmt, arg);
    raise_with_arg(GRAPHIC_FAILURE_EXN, copy_string(buffer));
}

void gr_fail()
{
    [lock lock];
    BOOL b = graphIsOpen;
    [lock unlock];
    if (b == NO)
    {
        [lock lock];
        if (data) {free(data); data = nil;}
        bitmap = nil;
        [timer invalidate];
        timer = nil;
        [lock unlock];
        queue = nil;
        grfail("graphic screen not opened", NULL);
    }
}

value gr_open_graph(value s)
{
    if (lock == nil) lock = [[NSLock alloc] init];
    [lock lock];
    if (graphIsOpen == YES) { [lock unlock]; return Val_unit; }
    graphIsOpen = YES;
    [lock unlock];
    int orx, ory;
    orx = 0;
    ory = 0;
    width = 600;
    height = 450;
    sscanf(String_val(s), "%dx%d+%d+%d", &width, &height, &orx, &ory);
    frame = NSMakeRect(orx, ory, width, height);
    title = @"Caml Graphics";
    @autoreleasepool {
#ifndef MACOS
        NSApp = [NSApplication sharedApplication];
        serverName = [[NSString alloc] initWithCString:getenv("CAMLSERVERNAME") encoding:NSISOLatin1StringEncoding];
#endif
        queue =[[NSMutableArray alloc] initWithCapacity:100];
        tailleImage = width*height*4;
        data = (unsigned char*)malloc(tailleImage);
        bitmap = [[NSBitmapImageRep alloc]
                  initWithBitmapDataPlanes:&data
                  pixelsWide:width
                  pixelsHigh:height
                  bitsPerSample:8
                  samplesPerPixel:4 hasAlpha:YES
                  isPlanar:NO
                  colorSpaceName:NSCalibratedRGBColorSpace
                  bytesPerRow:(4 * width)
                  bitsPerPixel:32];
        NSGraphicsContext* gc = [NSGraphicsContext graphicsContextWithBitmapImageRep:bitmap];
        [NSGraphicsContext setCurrentContext:gc];
        [gc setShouldAntialias:NO];
        [[NSColor whiteColor] set];
        NSRectFill(NSMakeRect(0,0,width,height));
        [[NSColor blackColor] set];
        [[NSFont fontWithName:@"Courier" size:12] set];
        strcpy(fontName, "Courier");
        needsRefresh = YES;
        newTitle = NO;
        timer = nil;
        x = 0;
        y = 0;
        fontSize = 12;
        lineWidth = 0;
        if (mcClient == nil) mcClient = [[MCClient alloc] init];
        [NSThread detachNewThreadSelector:@selector(mcThread) toTarget:mcClient withObject:nil];
    }
    return Val_unit;
}

value gr_close_graph(void)
{
    [lock lock];
    if (graphIsOpen == NO){[lock unlock]; return Val_unit;}
    graphIsOpen = NO;
    needsRefresh = NO;
    if (data)
    {
        [[NSColor whiteColor] set];
        NSRectFill(NSMakeRect(0,0,frame.size.width,frame.size.height));
        [mcProxy getImage: [NSData dataWithBytes: data length: tailleImage]];
        if (data) {free(data); data = nil;}
    }
    bitmap = nil;
    [mcProxy hideWindow]; // devrait vider la file des vnements du serveur
    [timer invalidate];
    timer = nil;
    free(data);
    queue = nil;
    [lock unlock];
    return Val_unit;
}

value gr_size_x(void)
{
    gr_fail();
    int w = frame.size.width;
    return Val_int(w);
}

value gr_size_y(void)
{
    gr_fail();
    int h = frame.size.height;
    return Val_int(h);
}

value gr_clear_graph(void)
{
    gr_fail();
    [lock lock];
    [[NSColor whiteColor] set];
    NSRectFill(NSMakeRect(0,0,frame.size.width,frame.size.height));
    [[NSColor colorWithCalibratedRed: red green: green blue: blue alpha: 1] set];
    needsRefresh = YES;
    [lock unlock];
    return Val_unit;
}

value gr_draw_line(value x1, value y1,value x2, value y2)
{
    gr_fail();
    float ax = Double_val(x1);
    float ay = Double_val(y1);
    float bx = Double_val(x2);
    float by = Double_val(y2);
    NSPoint p= NSMakePoint(ax, ay);
    NSPoint q= NSMakePoint(bx, by);
    [lock lock];
    @autoreleasepool {
        [NSBezierPath setDefaultLineWidth:lineWidth];
        [NSBezierPath strokeLineFromPoint:p toPoint:q];
    }
    needsRefresh = YES;
    [lock unlock];
    return Val_unit;
}

value gr_set_color(value c)
{
    gr_fail();
    int n = (int)Int_val(c);
    red = (n >> 16) / 255.;
    green = ((n >> 8) & 0xff) / 255.;
    blue = (n & 0xff) / 255.;
    [lock lock];
    @autoreleasepool {
        [[NSColor colorWithCalibratedRed: red green: green blue: blue alpha: 1.0] set];
    }
    [lock unlock];
    return Val_unit;
}

value gr_sigio_handler(void)
{
    return Val_unit;
}

value gr_moveto(value a, value b)
{
    gr_fail();
    x = Int_val(a);
    y = Int_val(b);
    return Val_unit;
}

value gr_current_point(void)
{
    gr_fail();
    value res = alloc_tuple(2);
    Field(res, 0) = Val_int(x);
    Field(res, 1) = Val_int(y);
    return res;
}

value gr_lineto(value a, value b)
{
    gr_fail();
    NSPoint p = NSMakePoint(x, y);
    x = Int_val(a);
    y = Int_val(b);
    NSPoint q = NSMakePoint(x, y);
    [lock lock];
    @autoreleasepool {
        [NSBezierPath setDefaultLineWidth:lineWidth];
        [NSBezierPath strokeLineFromPoint:p toPoint:q];
    }
    needsRefresh = YES;
    [lock unlock];
    return Val_unit;
}

value gr_set_window_title(value s)
{
    gr_fail();
    [lock lock];
    NSString* t = [NSString stringWithCString:String_val(s) encoding: NSUTF8StringEncoding];
    title = t;
    newTitle = YES;
    [lock unlock];
    return Val_unit;
}

value gr_plot(value x, value y)
{
    gr_fail();
    int ax = (int)Int_val(x);
    int ay = (int)Int_val(y);
    
    if (ax < 0 || ax >= width || ay < 0 || ay >= height) return Val_unit;
    
    unsigned char* buffer = [bitmap bitmapData];
    int k = (height - ay - 1)*width*4+ax*4;
    [lock lock];
    buffer[k++] = red * 255;
    buffer[k++] = green * 255;
    buffer[k++] = blue * 255;
    buffer[k] = 0xff;
    needsRefresh = YES;
    [lock unlock];
    return Val_unit;
}

value gr_point_color(value x, value y)
{
    gr_fail();
    int ax = (int)Int_val(x);
    int ay = (int)Int_val(y);
    if (ax < 0 || ax >= width || ay < 0 || ay >= height) return Val_int(16777215); // white pixel
    unsigned char* buffer = [bitmap bitmapData];
    int i = ((height - ay - 1) * width + ax) * 4;
    [lock lock];
    unsigned int r = buffer[i++];
    unsigned int g = buffer[i++];
    unsigned int b = buffer[i];
    unsigned int c = r * 65536 + g * 256 + b;
    [lock unlock];
    return Val_int(c);
}

value gr_set_line_width(value n)
{
    gr_fail();
    lineWidth = (int)Int_val(n);
    [NSBezierPath setDefaultLineWidth:lineWidth];
    return Val_unit;
}

value gr_draw_arc(value *argv, int argc)
{
    gr_fail();
    [lock lock];
    @autoreleasepool {
        [[NSGraphicsContext currentContext] saveGraphicsState];
        NSBezierPath* path = [NSBezierPath bezierPath];
        float cx = Int_val(argv[0]);
        float cy = Int_val(argv[1]);
        float r1 = Int_val(argv[2]);
        float r2 = Int_val(argv[3]);
        float a1 = Int_val(argv[4]);
        float a2 = Int_val(argv[5]);
        NSPoint c = NSMakePoint(cx,cy);
        [path moveToPoint: c];
        float r = r1 + r2;
        [path lineToPoint: NSMakePoint(cx + r * cos(pi*a1/180.), cy + r * sin(pi*a1/180.))];
        [path appendBezierPathWithArcWithCenter:c radius:r startAngle:a1 endAngle:a2];
        [path closePath];
        [path addClip];
        [[NSBezierPath bezierPathWithOvalInRect:NSMakeRect(cx - r1, cy -r2, 2 * r1, 2 * r2)] stroke];
        [[NSGraphicsContext currentContext] restoreGraphicsState];
    }
    needsRefresh = YES;
    [lock unlock];
    return Val_unit;
}

value gr_draw_ellipse(value a, value b, value r1, value r2)
{
    gr_fail();
    [lock lock];
    float cx = Int_val(a);
    float cy = Int_val(b);
    float rr1 = Int_val(r1);
    float rr2 = Int_val(r2);
    @autoreleasepool {
        [[NSBezierPath bezierPathWithOvalInRect:NSMakeRect(cx - rr1, cy -rr2, 2 * rr1, 2 * rr2)] stroke];
    }
    needsRefresh = YES;
    [lock unlock];
    return Val_unit;
}

value gr_draw_circle(value a, value b, value r)
{
    gr_fail();
    return gr_draw_ellipse(a, b, r, r);
}

value gr_fill_rect(value a, value b, value vw, value vh)
{
    gr_fail();
    [lock lock];
    int x = (int)Int_val(a);
    int y = (int)Int_val(b);
    int w = (int)Int_val(vw);
    int h = (int)Int_val(vh);
    @autoreleasepool {
        NSBezierPath* path = [NSBezierPath bezierPath];
        [path moveToPoint:NSMakePoint(x, y)];
        [path lineToPoint: NSMakePoint(x+w, y)];
        [path lineToPoint:NSMakePoint(x+w, y+h)];
        [path lineToPoint:NSMakePoint(x,y+h)];
        [path lineToPoint:NSMakePoint(x,y)];
        [path fill];
    }
    needsRefresh = YES;
    [lock unlock];
    return Val_unit;
}

value gr_fill_poly(value vect)
{
    gr_fail();
    int n_points, i;
    n_points = (int)Wosize_val(vect);
        @autoreleasepool {
        NSBezierPath* path = [NSBezierPath bezierPath];
        [path moveToPoint:NSMakePoint(Int_val(Field(Field(vect, 0), 0)), Int_val(Field(Field(vect, 0), 1)))];
        for(i = 1; i < n_points; i++)
            [path lineToPoint:NSMakePoint(Int_val(Field(Field(vect, i), 0)), Int_val(Field(Field(vect, i), 1)))];
        [path lineToPoint:NSMakePoint(Int_val(Field(Field(vect, 0), 0)), Int_val(Field(Field(vect, 0), 1)))];
        [lock lock];
        [path fill];
    }
    needsRefresh = YES;
    [lock unlock];
    return Val_unit;
}

value gr_fill_arc(value *argv, int argc)
{
    gr_fail();
    [lock lock];
    @autoreleasepool {
        [[NSGraphicsContext currentContext] saveGraphicsState];
        NSBezierPath* path = [NSBezierPath bezierPath];
        float cx = Int_val(argv[0]);
        float cy = Int_val(argv[1]);
        float r1 = Int_val(argv[2]);
        float r2 = Int_val(argv[3]);
        float a1 = Int_val(argv[4]);
        float a2 = Int_val(argv[5]);
        NSPoint c = NSMakePoint(cx,cy);
        [path moveToPoint: c];
        float r = r1 + r2;
        [path lineToPoint: NSMakePoint(cx + r * cos(pi*a1/180.), cy + r * sin(pi*a1/180))];
        [path appendBezierPathWithArcWithCenter:c radius:r startAngle:a1 endAngle:a2];
        [path closePath];
        [path addClip];
        [[NSBezierPath bezierPathWithOvalInRect:NSMakeRect(cx - r1, cy -r2, 2 * r1, 2 * r2)] fill];
        [[NSGraphicsContext currentContext] restoreGraphicsState];
    }
    needsRefresh = YES;
    [lock unlock];
    return Val_unit;
}

value gr_fill_ellipse(value a, value b, value r1, value r2)
{
    gr_fail();
    float cx = Int_val(a);
    float cy = Int_val(b);
    float rr1 = Int_val(r1);
    float rr2 = Int_val(r2);
    [lock lock];
    @autoreleasepool {
        [[NSBezierPath bezierPathWithOvalInRect:NSMakeRect(cx - rr1, cy -rr2, 2 * rr1, 2 * rr2)] fill];
    }
    needsRefresh = YES;
    [lock unlock];
    return Val_unit;
}

value gr_fill_circle(value a, value b, value r)
{
    gr_fail();
    return gr_fill_ellipse(a, b, r, r);
}

struct grimage
{
    final_fun f;                  /* Finalization function */
    int width, height;
    char* data;			//Pixels: 4 octets (R, G, B, Transp) au nombre de width * height
};

#define Grimage_wosize \
((sizeof(struct grimage) + sizeof(value) - 1) / sizeof(value))

#define Width_im(i) (((struct grimage *)(i))->width)
#define Height_im(i) (((struct grimage *)(i))->height)
#define Data_im(i) (((struct grimage *)(i))->data)

#define Max_image_mem 1000000

#define Max_young_wosize 256

static void gr_free_image(value im)
{
    free(Data_im(im));
}

value gr_make_image(value v)
{
    int i, j;
    
    gr_fail();
    int height = (int)Wosize_val(v);
    int width;
    if (height == 0)
    {
        width = 0;
    }
    else
    {
        width = (int)Wosize_val(Field(v, 0));
        for (i = 1; i < height; i++)
        {
            if (width != Wosize_val(Field(v, i)))
                grfail("make_image: lines of different lengths", NULL);
        }
    }
    value res= alloc_final(Grimage_wosize, gr_free_image, width * height, Max_image_mem);
    char* im = malloc(width * height * 4);
    char* data = im;
    for (i = 0; i <height; i++)
    {
        for (j = 0; j <width; j++)
        {
            int x = (int)Int_val(Field(Field(v, i), j));
            int r = x >> 16;
            int g = (x >> 8) & 0xff;
            int b = x & 0xff;
            if (x == -1)
            {
                *data++ = 0xff;
                *data++ = 0xff;
                *data++ = 0xff;
                *data++ = 0;
            }
            else
            {
                *data++ = r;
                *data++ = g;
                *data++ = b;
                *data++ = 0xff;
            }
        }
    }
    Width_im (res) = width;
    Height_im (res) = height;
    Data_im (res) = im;
    return res;
}

static value gr_alloc_int_vect(size)
mlsize_t size;
{
    value res;
    mlsize_t i;
    
    gr_fail();
    if (size <= Max_young_wosize) {
        res = alloc(size, 0);
    } else {
        res = alloc_shr(size, 0);
    }
    for (i = 0; i < size; i++) {
        Field(res, i) = Val_long(0);
    }
    return res;
}

value gr_dump_image(value image)
{
    int width, height, i, j;
    gr_fail();
    Push_roots(root, 2);
    
#define im root[0]
#define m root[1]
    
    width = Width_im(image);
    height = Height_im(image);
    im = image;
    m = gr_alloc_int_vect(height);
    for (i = 0; i < height; i++)
    {
        value v = gr_alloc_int_vect(width);
        modify(&Field(m, i), v);
    }
    char* data =Data_im(im);
    for (i = 0; i <height; i++)
    {
        for (j = 0; j <width; j++)
        {
            int r, g, b, t;
            r = *data++ & 0xff;
            g = *data++ & 0xff;
            b = *data++ & 0xff;
            t = *data++ & 0xff;
            if (t)
            {
                Field(Field(m, i), j) = Val_int((r << 16) + (g << 8) + b);
            }
            else
            {
                Field(Field(m, i), j) = Val_int(-1);
            }
        }
    }
    Pop_roots();
    return m;
#undef im
#undef m
}

void setPixelValue(unsigned char * buffer, int ax, int ay, int n, int w, int h)
{
    int k = (h - ay - 1)*w*4+ax*4;
    if (k < 0 || k > (w * h - 1) * 4) return;
    buffer[k++] = (n >> 16);
    buffer[k++] = ((n >> 8) & 0xff);
    buffer[k++] = (n & 0xff);
    buffer[k] = 0xff;
}

int getPixelValue(unsigned char * buffer, int ax, int ay, int w, int h)
{
    int k = (h - ay - 1)*w*4+ax*4;
    int r = buffer[k++];
    int g = buffer[k++];
    int b = buffer[k++];
    return (r << 16) + (g << 8) + b;
}

value gr_draw_image(value im, value vx, value vy)
{
    gr_fail();
    int w = Width_im(im);
    int h	= Height_im(im);
    unsigned char* imData = (unsigned char*)Data_im(im);
    int ax = (int)Int_val(vx);
    int ay = (int)Int_val(vy);
    [lock lock];
    int i;
    int j;
    for (i = 0;  i < w; i++) {
        for (j = 0;  j < h; j++) {
            if (i + ax < width && j + ay < height) {
                setPixelValue(data, i + ax, j + ay, getPixelValue(imData, i, j, w, h), width, height);
            }
        }
    }
    needsRefresh = YES;
    [lock unlock];
    return Val_unit;
}

value gr_create_image(value width, value height)
{
    gr_fail();
    int w, h;
    w = (int)Int_val(width);
    h = (int)Int_val(height);
    if (w < 0 || h < 0)
        grfail("create_image: width and height must be positive",NULL);
    value res = alloc_final(Grimage_wosize, gr_free_image, w * h, Max_image_mem);
    Width_im (res) = w;
    Height_im (res) = h;
    Data_im (res) = malloc(w * h * 4);
    return res;
}

value gr_blit_image(value im, value vx, value vy)
{
    gr_fail();
    int w = Width_im(im);
    int h = Height_im(im);
    unsigned char* imData = (unsigned char*)Data_im(im);
    int ax = (int)Int_val(vx);
    int ay = (int)Int_val(vy);
    if (ax < 0) ax = 0;
    if (ay < 0) ay = 0;
    [lock lock];
    
    int i;
    int j;
    for (i = 0;  i < w; i++) {
        for (j = 0;  j < h; j++) {
            if (i + ax < width && j + ay < height) {
                setPixelValue(imData, i, j, getPixelValue(data, i + ax, j + ay, width, height), w, h);
            }
        }
    }
    
    needsRefresh = YES;
    [lock unlock];
    return Val_unit;
}

value gr_draw_char(value c)
{
    gr_fail();
    char s[2];
    s[0] = Int_val(c);
    s[1] = '\0';
    @autoreleasepool {
        NSFont* font = [NSFont fontWithName: [NSString stringWithCString: fontName encoding: NSISOLatin1StringEncoding] size: fontSize];
        NSColor* color = [NSColor colorWithCalibratedRed:red green:green blue:blue alpha:1.0];
        NSDictionary* dico = [NSDictionary dictionaryWithObjects:
                              [NSArray arrayWithObjects:font, color, nil]
                                                         forKeys:
                              [NSArray arrayWithObjects:NSFontAttributeName, NSForegroundColorAttributeName, nil]];
        
        int width = [[NSString stringWithCString: s encoding: NSISOLatin1StringEncoding] sizeWithAttributes:dico].width;
        
        [[NSString stringWithCString:s encoding: NSISOLatin1StringEncoding] drawAtPoint:NSMakePoint(x, y) withAttributes:dico];
        x += width;
    }
    needsRefresh = YES;
    return Val_unit;
}

value gr_draw_string(value s)
{
    gr_fail();
    char* ss = String_val(s);
    @autoreleasepool {
        NSFont* font = [NSFont fontWithName: [NSString stringWithCString: fontName encoding: NSUTF8StringEncoding] size: fontSize];
        NSColor* color = [NSColor colorWithCalibratedRed:red green:green blue:blue alpha:1.0];
        NSDictionary* dico = [NSDictionary dictionaryWithObjects:
                              [NSArray arrayWithObjects:font, color, nil]
                                                         forKeys:
                              [NSArray arrayWithObjects:NSFontAttributeName, NSForegroundColorAttributeName, nil]];
        
        int width = [[NSString stringWithCString:ss encoding: NSUTF8StringEncoding] sizeWithAttributes:dico].width;
        
        [[NSString stringWithCString:ss encoding: NSUTF8StringEncoding] drawAtPoint:NSMakePoint(x, y) withAttributes:dico];
        x += width;
    }
    needsRefresh = YES;
    return Val_unit;
}

int text_length(char * str)
{
    NSUInteger lenA = (NSUInteger)strlen(str);
    NSUInteger len;
    @autoreleasepool {
        NSString* s = [[NSString alloc] initWithBytes:str length:lenA encoding: NSUTF8StringEncoding];
        len = s ? [s length] : -1;
    }
    return  (int)len;
}

value gr_text_length(value str)
{
    int i = text_length(String_val(str));
    if (i == -1) {invalid_argument("text_length"); return Val_unit;}
    return Val_int(i);
}

value gr_set_font(value s)
{
    gr_fail();
    @autoreleasepool {
        NSArray* availableFonts = [[NSFontManager sharedFontManager] availableFonts];
        NSEnumerator* enumerator = [availableFonts objectEnumerator];
        NSString* str;
        NSString* ss = [NSString stringWithCString:String_val(s) encoding: NSISOLatin1StringEncoding];
        while (str = [enumerator nextObject])
        {
            if ([str hasPrefix: ss])
            {
                [[NSFont fontWithName:str size:fontSize] set];
                strcpy(fontName, [str cStringUsingEncoding: NSISOLatin1StringEncoding]);
                return Val_unit;
            }
        }
    }
    return Val_unit;
}

value gr_set_text_size(value n)
{
    gr_fail();
    fontSize = Int_val(n);
    @autoreleasepool {
        NSFont* font = [NSFont fontWithName:[NSString stringWithCString: fontName encoding: NSUTF8StringEncoding] size: fontSize];
        [font set];
    }
    return Val_unit;
}

value gr_text_size(value s)
{
    gr_fail();
    NSSize size;
    @autoreleasepool {
        NSFont* font = [NSFont fontWithName: [NSString stringWithCString: fontName encoding: NSUTF8StringEncoding] size: fontSize];
        size = [[NSString stringWithCString: String_val(s) encoding: NSUTF8StringEncoding] sizeWithAttributes:
                       [NSDictionary dictionaryWithObjects: [NSArray arrayWithObjects:font, [NSColor blackColor], nil]
                                                   forKeys: [NSArray arrayWithObjects:NSFontAttributeName, NSForegroundColorAttributeName, nil]]];
    }
    value res = alloc_tuple(2);
    int width = size.width + 0.5;
    int height = size.height + 0.5;
    Field(res, 0) = Val_int(width);
    Field(res, 1) = Val_int(height);
    return res;
}

value gr_wait_event(value eventlist)
{
    gr_fail();
    int mask = 0;
    BOOL poll = NO;
    camlEvent cev;
    cev.mask = 0;
    cev.mouse_x = -1;
    cev.mouse_y = -1;
    cev.key = 0;
    while (Tag_val(eventlist) == 1) {
        switch (Tag_val(Field(eventlist, 0))) {
            case 0:
                mask |= MOUSEDOWN; break;
            case 1:
                mask |= MOUSEUP; break;
            case 2:
                mask |= KEYDOWN; break;
            case 3:
                mask |= MOUSEMOVE;
                break;
            case 4:
                poll = YES; break;
        }
        eventlist = Field(eventlist, 1);
    }
    if (poll)
    {
        [lock lock];
        int n = 0;
        if (queue != nil && (n =(int)[queue count]))
        {
            int i = 0;
            while (i < n)
            {
                MCEvent* me = [queue objectAtIndex:i];
                cev.mask = [me mask];
                cev.mouse_x = [me mouse_x];
                cev.mouse_y = [me mouse_y];
                cev.key = [me key];
                if ((cev.mask & KEYDOWN) != 0)
                {
                    break;
                }
                i++;
            }
        }
        [lock unlock];
    }
    else
    {
        while ((cev.mask & mask) == 0)
        {
            gr_fail();
            usleep(20);
            enter_blocking_section();
            leave_blocking_section();
            [lock lock];
            int n = 0;
            if (queue != nil && (n =(int)[queue count]))
            {
                int i = 0;
                while (i < n)
                {
                    MCEvent* me = [queue objectAtIndex:i];
                    cev.mask = [me mask];
                    cev.mouse_x = [me mouse_x];
                    cev.mouse_y = [me mouse_y];
                    cev.key = [me key];
                    if ((cev.mask & mask) != 0)
                    {
                        [queue removeObjectAtIndex:i];
                        break;
                    }
                    i++;
                }
            }
            [lock unlock];
        }
    }
    if (cev.mask & MOUSEMOVE)
    {
        [mcProxy getImage: [NSData dataWithBytes: data length: tailleImage]];
    }
    value res = alloc_tuple(5);
    Field(res, 0) = Val_int(cev.mouse_x);
    Field(res, 1) = Val_int(cev.mouse_y);
    Field(res, 2) = Val_bool(cev.mask & (MOUSEDOWN | MOUSEMOVE));
    Field(res, 3) = Val_bool(cev.mask & KEYDOWN);
    Field(res, 4) = Val_int(cev.key & 0xFF);
    return res;
}

value gr_sound(value freq, value duration)
{
    gr_fail();
    NSBeep(); //  faire...
    return Val_unit;
}
